'******************************************************************************
'*  trs80.spin - Demo mode for the TRS80 video and render cogs
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id: cgenie.spin,v 1.5 2010-06-05 05:59:21 pm Exp $
'******************************************************************************
CON
	_xinfreq		= 5_000_000		' Quartz is 5MHz
	_clkmode		= xtal1 + pll16x	' System clock is 80MHz

' We need a Z80 to run
#define CPU_Z80

' Memory access dispatch tables for CGENIE in qz80
#define CGENIE

' If you want to use VGA out, define VGA and disable the #define TV above
#define VGA

'#define TriBladeProp
'#define RamBladeProp
#define DracBladeProp

' Uncomment to have debug code inserted
'#define DEBUG

' Uncomment to emulate interrupts (NMI and IRQ)
'#define INTERRUPTS

' Uncomment to count cycles
'#define COUNT_C

' Uncomment to count R register
#define COUNT_R


#ifdef TriBladeProp
' serial I/O pins on the TriBladeProp #2
	SI			= 31
	SO			= 30
' SD pins on the TriBladeProp #2
	doPin			= 9
	clkPin			= 28
	diPin			= 8
	csPin			= 14
' TV out pin
	tvPin			= 14
' Keyboard pins
	kbDpin			= 26
	kbCpin			= 27
' audio pin
	auPin			= 16	' ???
#endif

#ifdef RamBladeProp
' serial I/O pins on the RamBladeProp
	SI			= 23
	SO			= 22
' microSD pins on RamBladeProp
	doPin			= 24
	clkPin			= 26
	diPin			= 25
	csPin			= 19
' TV out pin
	tvPin			= 14
' Keyboard pins
	kbDpin			= 26
	kbCpin			= 27
' audio pin
	auPin			= 22	' ??? instead of SO?
#endif

#ifdef DracBladeProp
' serial I/O pins on the DracBladeProp
	SI			= 31
	SO			= 30
' SD pins on the DracBladeProp
	doPin			= 12
	clkPin			= 13
	diPin			= 14
	csPin			= 15
' TV out pin
	tvPin			= 16
' Keyboard pins
	kbDpin			= 26
	kbCpin			= 27
' audio pin
	auPin			= 24	' SIO2 TX
#endif

VAR
	byte	pins[10]
	long	params[6]

OBJ
	io	: "io"
	qz80	: "qz80"

PUB     main : okay
	pins[ 0] := SI					' currently unused
	pins[ 1] := SO
	pins[ 2] := doPin
	pins[ 3] := clkPin
	pins[ 4] := diPin
	pins[ 5] := csPin
	pins[ 6] := kbDpin
	pins[ 7] := kbCpin
	pins[ 8] := auPin
	pins[ 9] := tvPin

	okay := io.start(@pins)
	if okay < 0
		repeat

	params[0] := io.get_iocmd_ptr
	params[1] := io.get_irq				' interrupts not used
	params[2] := io.get_videoram
	params[3] := io.get_startaddr
	params[4] := io.get_colourram
	params[5] := io.get_fontram

	qz80.start(@params, 0)

''      MIT LICENSE
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}
