CON
	' 8085 flag registers
	CF			=	%00000001	' carry
	NF			=	%00000010	' negative
	VF			=	NF
	PF			=	%00000100	' parity (also overflow)
	X3			=	%00001000	' undocumented x
	HF			=	%00010000	' auxiliary (half) carry
	X5			=	%00100000	' undocumented y
	ZF			=	%01000000	' zero
	SF			=	%10000000	' sign
	
	io_cmd_out		=	$01		' command to write to an i/o port
	io_cmd_in		=	$02		' command to read from an i/o port
	io_cmd_break		=	$03		' command to execute a break (halt)
	io_cmd_kbd		=	$04		' command to read the keyboard matrix
	io_cmd_get_sid		=	$05		' get serial input device status (SID)
	io_cmd_set_sod		=	$06		' set serial output device status (SOD)

	IM_SID          	=	$80
	IM_I75          	=	$40
	IM_I65          	=	$20
	IM_I55          	=	$10
	IM_IE           	=	$08
	IM_M75          	=	$04
	IM_M65          	=	$02
	IM_M55          	=	$01

	ADDR_TRAP       	=	$0024
	ADDR_RST55      	=	$002c
	ADDR_RST65      	=	$0034
	ADDR_RST75      	=	$003c
	ADDR_INTR       	=	$0038
	ADDR_RSTV		=	$0040

#ifdef GAMEBOY
#define PAGED_MEMORY
#endif

VAR
	long	cog

PUB start(params) : okay
	stop
	io_command := long[params][0]
#ifdef BANKED_MEM
	bankbase_ptr := long[params][0] + 4
#endif
#ifdef INTERRUPTS
	irq_ptr := long[params][1]
#endif
#ifdef GAMEBOY
	' other pointers required?
	screen_ptr := long[params][2] - $8000
#endif GAMEBOY
	COGINIT(COGID, @entry, 0)

PUB stop
	if cog
		COGSTOP(cog~~ - 1)

DAT
	org	0
entry

#ifdef TriBladeProp
'The TriBladeProp Blade #2 driver presets the latch to enable SRAM -CE(U23)=0 and hold LE=0
datax		mov     outa, ram_latched		' -WE=1 -OE=1, all other bits=0
outx		mov     dira, ram_dir_read		' setup direction to read
		jmp	#fetch				' contains hub RAM address for ea (overwrites jmp)
#endif TriBladeProp

#ifdef RamBladeProp
'The RamBladeProp needs this?
datax		mov     outa, #0
outx		jmp	#fetch
#endif RamBladeProp

#ifdef DracBladeProp
outx		jmp	#fetch				' contains address masks for ram_address setup (overwrites jmp)
oldx		long	-1				' for RAM latch speed up
#endif DracBladeProp

io_command	long    0-0				' io command address
io_break	long	io_cmd_break << 24 | @@@C_reg << 8

#ifdef INTERRUPTS
irq_ptr		long	0-0				' interrupt request address
do_int		long	@@@do_interrupt
HALT		long	0
#endif INTERRUPTS

table_00	long	@@@opcodes_00
debug_on	long	0

#ifdef PAGED_MEMORY
rd_memory_ptr	long	@@@rd_memory
wr_memory_ptr	long	@@@wr_memory
screen_ptr	long	0-0
#endif

#ifdef BANKED_MEM
bankbase_ptr	long	0-0
banklimit	long	$c000
bankbase	long	0
#endif BANKED_MEM

context_ptr	long	@@@C_reg			' C register address
minus1		long	$ffffffff
low_word	long	$0000ffff
d1		long	1<<9				' add 1 to destination

C		long	0
B		long	0
E		long	0
D		long	0
L		long	0
H		long	0
F		long	ZF				' reset value is zero flag set
A		long	0
IM		long	0				' interrupt mask register
IM_TRAP		long	0				' interrupt mask register at most recent TRAP
SP		long	0				' stack pointer
PC		long	0				' program counter

'***************************************************************************************************************
' Fetch opcode and dispatch through table_00
'
fetch
#ifdef INTERRUPTS
		rdlong	tmp, irq_ptr		WZ
	if_nz	mov	lmm_pc, do_int
	if_nz	call	#lmm_next
fetch_0
#endif INTERRUPTS
		call	#rd_opcode
		mov	opcode, alu
		shl	alu, #2
		add	alu, table_00
		rdlong	vector, alu
		jmp	vector


'***************************************************************************************************************
' Jump to next address in vector
'
jmp_vector
		shr	vector, #9
		jmp	vector

#ifdef PAGED_MEMORY
mem_page_1k
		mov	outx, ea
		shr	outx, #10			' discard 10 bits of the address
mem_table	add	outx, 0-0			' source is patched with the table address
		rdbyte	outx, outx			' fetch memory access handler
		jmp	outx				' and jump there
rd_rom
		jmp	#rd_ram				' same as RAM
rd_nop
		mov	alu, #$ff			' read nothing but FF
		jmp	#rd_byte_ret
rd_i_o
		' TODO: read from memory mapped i/o
		mov	alu, #$ff			' not yet implemented
		jmp	#rd_byte_ret
rd_scr
		mov	outx, ea
		add	outx, screen_ptr		' read from screen memory
		rdbyte	alu, outx
		jmp	#rd_byte_ret
rd_kbd
		mov	t1, #io_cmd_kbd			' read from keyboard matrix
		shl	t1, #16
		or	t1, ea				' set address in io_command long
		shl	t1, #8
		call	#rd_port_cmd
		jmp	#rd_byte_ret
wr_i_o
		' TODO: write to memory mapped i/o
		jmp	#wr_byte_ret			' not yet implemented
wr_scr
		mov	outx, ea
		add	outx, screen_ptr
		wrbyte	alu, outx
		jmp	#wr_byte_ret
#endif PAGED_MEMORY

'***************************************************************************************************************
'
'	Memory access functions (TriBladeProp)
'
'***************************************************************************************************************
#ifdef TriBladeProp
'                                          +-------- -CS (U26 FLASH)   -+                                  
'                                          |+------- -CS (J22 microSD)  +- Latched pins...                 
'                                          ||+------ -RST3              |    passes when LE=1              
'             (P30) SO ---+                |||+----- -RST1              |    latches on  LE low going edge 
'             (P31) SI --+|                ||||+---- -CE (U24 SRAM-hi)  |    latched on  LE=0              
'                        ||+---- -WE(SDA)  |||||+--- -CE (U23 SRAM-lo)  |                                  
'                        |||+--- -OE(SCL)  ||||||+-- A20                |                                  
'                        ||||+--  LE       |||||||+- A19               -+                                  
'                        |||||             ||||||||                                                        
'          P31...P0 --> %00xx0_xxxxxxxxxxx_xxxxxxxx_xxxxxxxx  
ram_latched	long	%00110_00000000000_00000000_00000000	' -WE=1 -OE=1 (LE=0)                               
ram_read	long	%00100_00000000000_00000000_00000000	' -WE=1 -OE=0 (LE=0)    
ram_write	long	%00010_00000000000_00000000_00000000	' -WE=0 -OE=1 (LE=0)
ram_dir_read	long	%00110_11111111111_11111111_00000000	' outputs WE, OE, A0-18, inputs D0-7
ram_dir_write	long	%00110_11111111111_11111111_11111111	' outputs WE, OE, A0-18, D0-7
ram_dir_input	long	%00000_00000000000_00000000_00000000	' all inputs (for Ram Disk access)
'                              +---------+ +------+ +------+
'                              A18......A8 A7....A0 D7....D0
'                                          Q7....Q0
'
'***************************************************************************************************************
' Read a byte from RAM[HL] into the alu
'
rd_byte_hl
		mov	ea, H
		shl	ea, #8
		or	ea, L
'***************************************************************************************************************
' Read a byte from RAM[ea] into the alu
'
rd_byte
#ifdef BANKED_MEM
		cmp	ea, banklimit		WC
	if_c	add	ea, bankbase
#endif BANKED_MEM
#ifdef PAGED_MEMORY
		movs	mem_table, #rd_memory_ptr
		jmp	#mem_page_1k
rd_ram
#endif PAGED_MEMORY
		mov	outx, ea
		shl	outx, #8			' shift address 8 bits
		or	outx, ram_read			' add -WE=1 -OE=0
		mov	outa, outx			' pc (shifted 8 bits + -WE=1 -OE=0)
		nop					' delay for sram access time
		mov     alu, ina			' read SRAM
		and     alu, #$ff			' extract 8 bits
rd_byte_ret
rd_byte_hl_ret
		ret

'***************************************************************************************************************
' Write a byte from the alu into RAM[ea]
'
wr_byte
#ifdef BANKED_MEM
		cmp	ea, banklimit		WC
	if_c	add	ea, bankbase
#endif BANKED_MEM
#ifdef PAGED_MEMORY
		movs	mem_table, #wr_memory_ptr	' dispatch according to the wr_memory table
		jmp	#mem_page_1k			' in 1k pages
wr_ram
#endif PAGED_MEMORY
		mov	outx, ea
		shl	outx, #8			' shift address 8 bits
		or	outx, ram_write			' add -WE=0 -OE=1
		mov	datax, alu
		and	datax, #$ff			' ensure upper bytes=0
		or	outx, datax			' add data bits to be written
		mov	outa, outx			' output -WE=0 -OE=1, address bits, data bits
		mov	dira, ram_dir_write		' set data bits to outputs
		or      outa, ram_latched		' set -WE=0->1 -OE=1->1  (end write)
		mov     dira, ram_dir_read		' set data bits to inputs
		mov     outa, ram_read			' set -WE=1 -OE=0
#ifdef PAGED_MEMORY
wr_rom
wr_nop
#endif PAGED_MEMORY
wr_byte_ret
		ret
#endif TriBladeProp





'***************************************************************************************************************
'
'	Memory access functions (RamBladeProp)
'
'***************************************************************************************************************
#ifdef RamBladeProp
' RamBlade Hardware Access:
'       P0-18  = A0-18   SRAM address
'       P19    = -CS     microSD
'       P20    = -CE     SRAM & enable eeprom
'       P21    = -WE     SRAM & enable eeprom
'       P22    = SO      \ UART
'       P23    = SI      /
'       P24-31 = D0-7    SRAM \ * see below
'                             | P28 = SCL eeprom
'                             | P29 = SDA eeprom
'                             | P30 = prog SO
'                             / P31 = prog SI
'
'         (microSD) DO ---------+
'         (microSD) DI --------+| +--------  SI (P23)
'         (microSD) CLK-------+|| |+-------  SO (P22)
'                             ||| ||
'             (P28) SCL ----+ ||| ||+------ -WE (SRAM & enable eeprom)
'             (P29) SDA ---+| ||| |||+----- -CE (SRAM & enable eeprom)
'             (P30) SO ---+|| ||| ||||+---- -CS (microSD)
'             (P31) SI --+||| ||| |||||
'                        |||| ||| |||||
'          P31...P0 --> %xxxxxxxx_00xxx_xxxxxxxxxxxxxxxxxxx
'       10K pullups -->    ** ***   ***
'                        +------+       +-----------------+
'                        D7....D0       A18..............A0
ram_dir_read    long    %00000000_00010_1111111111111111111 ' outputs CE, A0-18           (sram read)
ram_dir_write   long    %11111111_00000_1111111111111111111 ' outputs A0-18, D0-7         (sram disabled)
ram_dir_wstrobe long    %11111111_00110_1111111111111111111 ' outputs CE, WE, A0-19, D0-7 (write strobe)
'
'***************************************************************************************************************
' Read a byte from RAM[HL] into the alu
'
rd_byte_hl
		mov	ea, H
		shl	ea, #8
		or	ea, L
'***************************************************************************************************************
' Read a byte from RAM[ea] into the alu
'
rd_byte
#ifdef BANKED_MEM
		cmp	ea, banklimit		WC
	if_c	add	ea, bankbase
#endif BANKED_MEM
#ifdef PAGED_MEMORY
		movs	mem_table, #rd_memory_ptr
		jmp	#mem_page_1k
rd_ram
#endif PAGED_MEMORY
		mov	OUTA, ea			' apply the read address incl -CE=0 -WE=0
		mov	DIRA, ram_dir_read		' set outputs A0-18, CE
		nop					' sram delay required
		mov	alu, INA			' read byte from SRAM
		shr	alu, #24			' bits 31..24 to 7..0
		mov	DIRA, #0			' tri state
rd_byte_ret
rd_byte_hl_ret
		ret

'***************************************************************************************************************
' Write a byte from the alu into RAM[ea]
'
wr_byte
#ifdef BANKED_MEM
		cmp	ea, banklimit		WC
	if_c	add	ea, bankbase
#endif BANKED_MEM
#ifdef PAGED_MEMORY
		movs	mem_table, #wr_memory_ptr
		jmp	#mem_page_1k
wr_ram
#endif PAGED_MEMORY
		shl	alu, #24			' data byte in bits 31..24
		or	alu, ea				' data in 31..24, address in 18..0
		or	OUTA, alu
		mov	DIRA, ram_dir_write		' set outputs A0-18
		mov	DIRA, ram_dir_wstrobe		' -CE=-WE= 1 -> 0       \ ce + we pulse
		shr	alu, #24			' data byte in bits 7..0 | delay
		mov	DIRA, ram_dir_write		' address + data only   /
		mov	DIRA, #0			' tri state
#ifdef PAGED_MEMORY
wr_rom
wr_nop
#endif PAGED_MEMORY
wr_byte_ret
		ret
#endif RamBladeProp





'***************************************************************************************************************
'
'	Memory access functions (DracBladeProp)
'
'***************************************************************************************************************
#ifdef DracBladeProp
'****************** Dr_Acula single propeller driver **************************
' Three latches to access the ram chip - see main program for spin version
' uses variables 'ea' and 'alu'. skip banked memory and checking for out of
' bounds address will need two nop instructions in /rd and /wr to be sure 
'
dir_138_read	long    %00000000_00000000_00001111_00000000	' for reads so data lines are tristate till the read
dir_138_write	long    %00000000_00000000_00001111_11111111	' for writes so data lines are outputs
address_low	long    %00000000_00000000_00000100_00000000	' low address latch = xxxx010x and gate low xxxxxxx0
address_mid	long    %00000000_00000000_00000110_00000000	' middle address latch = xxxx011x and gate low xxxxxxx0
address_high	long    %00000000_00000000_00001000_00000000	' highe address latch = xxxx100x and gate low xxxxxxx0
write_enable	long    %00000000_00000000_00000010_00000000	' write enable xxxx001x and gate low xxxxxxx0
'
'***************************************************************************************************************
' Read a byte from RAM[HL] into the alu
'
rd_byte_hl
		mov	ea, H
		shl	ea, #8
		or	ea, L
'***************************************************************************************************************
' Read a byte from RAM[ea] into the alu
'
rd_byte
#ifdef PAGED_MEMORY
		movs	mem_table, #rd_memory_ptr
		jmp	#mem_page_1k
rd_ram
#endif PAGED_MEMORY
		call	#ram_address			' sets up the latches with the correct ram address
		mov	DIRA, dir_138_read		' for reads so P0-P7 tristate till do read
		mov	OUTA, #$0			' set 000 low 138 read, low gate and P0-P7 are inputs
		nop					' none does not work, 1 does work
		mov	alu, INA			' read SRAM
		and	alu, #$ff			' extract 8 bits
		or	OUTA, #%1_00000000		' set the gate high again
		mov	DIRA, #0			' tristate all pins
rd_byte_ret
rd_byte_hl_ret
		ret

'***************************************************************************************************************
' Write a byte from the alu into RAM[ea]
'
wr_byte
#ifdef PAGED_MEMORY
		movs	mem_table, #wr_memory_ptr
		jmp	#mem_page_1k
wr_ram
#endif PAGED_MEMORY
        	call	#ram_address			' sets up the latches with the correct ram address
        	mov	outx, alu			' get the byte to output
        	and	outx, #$ff			' ensure upper bytes=0
        	or	outx, write_enable		' or with correct 138 address which is the same as latchdirection2
        	mov	OUTA, outx			' send it out
        	nop					' NOP  needs at least one due to 138 propogation delay
        	or	OUTA, #%1_00000000		' set it high again
        	mov	DIRA, #0			' tristate all pins
#ifdef PAGED_MEMORY
wr_rom
wr_nop
#endif PAGED_MEMORY
wr_byte_ret
		ret

'***************************************************************************************************************
' Sets up the DracBlade RAM latches.
' BANKED_MEM undefined: assumes high latch A16-A18 low so only accesses 64K of RAM
' BANKED_MEM defined: adds current bank base for addresses below banklimit
ram_address
#ifdef BANKED_MEM
		cmp	ea, banklimit		WC
	if_c	add	ea, bankbase
#endif BANKED_MEM
		mov	DIRA, dir_138_write		' set up the pins for programming latch chips
		mov	outx, ea			' get the address into a temp variable
		and	outx, #$ff			' mask the low byte
		or	outx, address_low		' or with 138 low address
		mov	OUTA, outx			' send it out with gate low
		xor	oldx, ea
		andn	oldx, #$ff
		or	OUTA, #%1_00000000		' set gate high again and repeat for the middle byte
		tjz	oldx, #:done
		mov	outx, ea			' get the address into a temp variable
		shr	outx, #8			' shift right by 8 places
		and	outx, #$ff			' mask the middle byte
		or	outx, address_mid		' or with 138 middle address
		mov	OUTA, outx			' send it out with gate low
		or	OUTA, #%1_00000000		' set gate high again
#ifdef BANKED_MEM
		mov	outx, ea
		andn	oldx, low_word
		tjz	oldx, #:done
		shr	outx, #16			' shift right by 16 places
		and	outx, #$07			' mask 3 bits in the high byte
		or	outx, address_high		' or with 138 high address
		mov	OUTA, outx			' send it out with gate low
		or	OUTA, #%1_00000000		' set gate high again
#endif BANKED_MEM
:done		mov	oldx, ea
ram_address_ret
		ret
#endif DracBladeProp





'***************************************************************************************************************
'
' Read a byte from address [PC++] into alu
'
rd_opcode
		mov	ea, PC
		add	PC, #1
		and	PC, low_word
		call	#rd_byte
rd_opcode_ret
		ret

'***************************************************************************************************************
'
' Read a word from address [PC++] into alu
'
rd_opword
		mov	ea, PC
		add	PC, #2
		and	PC, low_word
		jmp	#rd_word

'******************************************************************************
'
' Pop a word off the stack
'
pop_alu
		mov	ea, SP
		add	SP, #2
		and	SP, low_word

'***************************************************************************************************************
'
' Read a word from address [ea] into alu
'
' clobbers: t1
'
rd_word
		call	#rd_byte
		mov	t1, alu
		add	ea, #1
		and	ea, low_word
		call	#rd_byte
		shl	alu, #8
		or	alu, t1
		sub	ea, #1
rd_word_ret
rd_opword_ret
pop_alu_ret
		ret

'******************************************************************************
'
' Push a word to the stack
'
push_alu
		sub	SP, #2
		and	SP, low_word
		mov	ea, SP

'***************************************************************************************************************
'
' Write a word from alu to address [ea]
'
wr_word
		call	#wr_byte
		add	ea, #1
		and	ea, low_word
		ror	alu, #8
		call	#wr_byte
wr_word_ret
push_alu_ret
		ret

'***************************************************************************************************************
' Write a byte from the alu to an I/O port
'
wr_port
#ifdef TriBladeProp
		mov     dira, ram_dir_input		' pass I/O to another cog
#endif TriBladeProp
		mov	t1, #io_cmd_out
		shl	t1, #16
		or	t1, ea
		shl	t1, #8
		or	t1, alu
		wrlong	t1, io_command
:wait		rdlong	t1, io_command			'...and wait for it to be completed.
		shr	t1, #24		WZ, NR
	if_nz	jmp	#:wait
#ifdef TriBladeProp
		mov     dira, ram_dir_read		' take I/O from another cog
#endif TriBladeProp
#ifdef BANKED_MEM
		rdlong	bankbase, bankbase_ptr		' bankbase may have been modified
#endif BANKED_MEM
		neg	oldx, oldx
wr_port_ret
		ret

'***************************************************************************************************************
' Read a byte from an I/O port into the alu
'
rd_port
		mov	t1, #io_cmd_in
		shl	t1, #16
		or	t1, ea
		shl	t1, #8
rd_port_cmd
#ifdef TriBladeProp
		mov     dira, ram_dir_input		' pass I/O to another cog
#endif TriBladeProp
		wrlong	t1, io_command
:wait		rdlong	t1, io_command			'...and wait for it to be completed.
		shr	t1, #24		WZ, NR
	if_nz	jmp	#:wait
#ifdef TriBladeProp
		mov     dira, ram_dir_read		' take I/O from another cog
#endif TriBladeProp
		mov	alu, t1
		and	alu, #$ff
		neg	oldx, oldx
rd_port_cmd_ret
rd_port_ret
		ret

'***************************************************************************************************************
' Push a register from alu and jump to fetch
'
push_reg
		movs	push_alu_ret, #fetch
		jmp	#push_alu

'***************************************************************************************************************
' Transfer B register to alu
'
b_to_alu
		mov	alu, B
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer C register to alu
'
c_to_alu
		mov	alu, C
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer D register to alu
'
d_to_alu
		mov	alu, D
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer E register to alu
'
e_to_alu
		mov	alu, E
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer H register to alu
'
h_to_alu
		mov	alu, H
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer L register to alu
'
l_to_alu
		mov	alu, L
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer (HL) to alu
'
m_to_alu
		movs	rd_byte_hl_ret, #jmp_vector
		jmp	#rd_byte_hl

'***************************************************************************************************************
' Transfer A register to alu
'
a_to_alu
		mov	alu, A
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer port[imm8] to alu
'
port_to_a
		call	#rd_opcode
		mov	ea, alu
		movs	rd_port_ret, #alu_to_a
		jmp	#rd_port

'***************************************************************************************************************
' Transfer BC register pair to alu
'
bc_to_alu
		mov	alu, B
		shl	alu, #8
		or	alu, C
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer DE register pair to alu
'
de_to_alu
		mov	alu, D
		shl	alu, #8
		or	alu, E
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer HL register pair to alu
'
hl_to_alu
		mov	alu, H
		shl	alu, #8
		or	alu, L
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer AF register pair to alu
'
af_to_alu
		mov	alu, A
		shl	alu, #8
		or	alu, F
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer SP register to alu
'
sp_to_alu
		mov	alu, SP
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer BC register pair to ea
'
bc_to_ea
		mov	ea, B
		shl	ea, #8
		or	ea, C
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer DE register pair to ea
'
de_to_ea
		mov	ea, D
		shl	ea, #8
		or	ea, E
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
' Transfer HL register pair to ea
'
hl_to_ea
		mov	ea, H
		shl	ea, #8
		or	ea, L
		shr	vector, #9
		jmp	vector

'***************************************************************************************************************
'
' NOTE: the following 8 entries must be in the sequence B, C, D, E, H, L, M, A and consist of 2 instructions
'
'***************************************************************************************************************
' Transfer alu to B register
'
alu_to_b
		mov	B, alu
		jmp	#fetch

'***************************************************************************************************************
' Transfer alu to C register
'
alu_to_c
		mov	C, alu
		jmp	#fetch

'***************************************************************************************************************
' Transfer alu to D register
'
alu_to_d
		mov	D, alu
		jmp	#fetch

'***************************************************************************************************************
' Transfer alu to E register
'
alu_to_e
		mov	E, alu
		jmp	#fetch

'***************************************************************************************************************
' Transfer alu to H register
'
alu_to_h
		mov	H, alu
		jmp	#fetch

'***************************************************************************************************************
' Transfer alu to L register
'
alu_to_l
		mov	L, alu
		jmp	#fetch

'***************************************************************************************************************
' Transfer alu to (HL)
'
alu_to_m
		movs	wr_byte_ret, #fetch
		jmp	#wr_byte

'***************************************************************************************************************
' Transfer alu to A register
'
alu_to_a
		mov	A, alu
		jmp	#fetch


'***************************************************************************************************************
' Transfer alu to port[imm8]
'
a_to_port
		call	#rd_opcode
		mov	ea, alu
		mov	alu, A
		movs	wr_port_ret, #fetch
		jmp	#wr_port

'***************************************************************************************************************
' Pop BC register pair off the stack
'
pop_bc
		call	#pop_alu
		' fall through
'***************************************************************************************************************
' Transfer alu to BC register pair
'
alu_to_bc
		mov	B, alu
		shr	B, #8
		mov	C, alu
		and	C, #$ff
		jmp	#fetch

'***************************************************************************************************************
' Pop DE register pair off the stack
'
pop_de
		call	#pop_alu
		' fall through
'***************************************************************************************************************
' Transfer alu to DE register pair
'
alu_to_de
		mov	D, alu
		shr	D, #8
		mov	E, alu
		and	E, #$ff
		jmp	#fetch

'***************************************************************************************************************
' Pop HL register pair off the stack
'
pop_hl
		call	#pop_alu
		' fall through
'***************************************************************************************************************
' Transfer alu to HL register pair
'
alu_to_hl
		mov	H, alu
		shr	H, #8
		mov	L, alu
		and	L, #$ff
		jmp	#fetch

'***************************************************************************************************************
' Pop AF register pair off the stack
'
pop_af
		call	#pop_alu
		' fall through
'***************************************************************************************************************
' Transfer alu to AF register pair
'
alu_to_af
		mov	A, alu
		shr	A, #8
		mov	F, alu
		and	F, #$ff
		jmp	#fetch

'***************************************************************************************************************
' Transfer alu to SP register
'
alu_to_sp
		mov	SP, alu
		jmp	#fetch

'******************************************************************************
'
' Return from subroutine
'
do_ret
		call	#pop_alu
		' fall through
'***************************************************************************************************************
' Transfer alu to PC register
'
alu_to_pc
		mov	PC, alu
		jmp	#fetch

'***************************************************************************************************************
' Fetch an imm8 value at memory (PC++)
'
imm8_to_alu
		movs	rd_opcode_ret, #jmp_vector
		jmp	#rd_opcode

'***************************************************************************************************************
' Fetch an imm16 value at memory (PC++)
'
imm16_to_alu
		movs	rd_opword_ret, #jmp_vector
		jmp	#rd_opword

'***************************************************************************************************************
' Fetch an imm16 value at memory (PC++) into ea
'
imm16_to_ea
		call	#rd_opword
		mov	ea, alu
		jmp	#jmp_vector
'***************************************************************************************************************
' Read a memory byte (ea) into alu
'
rd_byte_ea
		movs	rd_byte_ret, #jmp_vector
		jmp	#rd_byte

'***************************************************************************************************************
' Write alu to a memory byte (ea)
'
wr_byte_ea
		movs	wr_byte_ret, #fetch
		jmp	#wr_byte

'***************************************************************************************************************
' Read a memory word (ea), (ea+1) into alu
'
rd_word_ea
		movs	rd_word_ret, #jmp_vector
		jmp	#rd_word

'***************************************************************************************************************
' Write alu to a memory word (ea), (ea+1)
'
wr_word_ea
		movs	wr_word_ret, #fetch
		jmp	#wr_word

'***************************************************************************************************************
' Test a flag vector[16..9] in the F register
'
test_flag_1
		mov	tmp, vector
		shr	tmp, #18
		and	tmp, F		WZ
		jmp	#jmp_vector

'***************************************************************************************************************
' Test a flag vector[16..9] in the F register
'
test_flag_0
		mov	tmp, vector
		shr	tmp, #18
		andn	tmp, F		WZ
		jmp	#jmp_vector

'******************************************************************************
'
' Call a restart vector
'
rst
		mov	alu, opcode
		and	alu, #$38

'******************************************************************************
'
' Call subroutine
'
do_call
		xor	alu, PC			' swap alu and PC
		xor	PC, alu
		xor	alu, PC
		movs	push_alu_ret, #fetch
		jmp	#push_alu

'******************************************************************************
'
' RET if the condition is true
'
ret_cond
	if_nz	jmp	#do_ret
		jmp	#fetch

'******************************************************************************
'
' JP if the condition is true
'
jmp_cond
		call	#rd_opword
	if_nz	mov	PC, alu
		jmp	#fetch

'******************************************************************************
'
' CALL if the condition is true
'
call_cond
		call	#rd_opword
	if_nz	jmp	#do_call
		jmp	#fetch

'******************************************************************************
'
' Increment 16 bit register pair
'
inc16
		add	alu, #1
		and	alu, low_word
		jmp	#jmp_vector
'******************************************************************************
'
' Decrement 16 bit register pair
'
dec16
		sub	alu, #1
		and	alu, low_word
		jmp	#jmp_vector

'***************************************************************************************************************
'
' Compute an 8 bit increment and set the flags
'
alu_inc
		movs	szhp_flags_cmp, #$00		' compare auxiliary value
		andn	F, #NF				' reset negative flag
		add	alu, #1
szhp_flags
		mov	aux, alu
		and	aux, #$0f
szhp_flags_cmp	cmp	aux, #$00/$0f	WZ		' carry from bit 3 to 4?
		muxz	F, #HF				' set or clear half carry
		and	alu, #$ff	WZ, WC		' result is zero?
		muxz	F, #ZF				' set or clear zero
		muxnc	F, #PF				' clear P flag on parity
		test	alu, #$80	WC		' bit #7 set?
		muxc	F, #SF				' set or clear sign
		jmp	#jmp_vector

'***************************************************************************************************************
'
' Compute an 8 bit decrement and set the flags
'
alu_dec
		movs	szhp_flags_cmp, #$0f		' compare auxiliary value
		or	F, #NF				' set negative flag
		sub	alu, #1				' decrement alu
		jmp	#szhp_flags

'***************************************************************************************************************
'
' Compute sign, zero, auxiliary, overflow and carry flags after an addition or subtraction
'
szhpc_flags
		xor	aux, tmp			' aux = result ^ operand ^ argument
		xor	aux, alu
		test	aux, #HF	WC		' H flag change?
		muxc	F, #HF
		test	aux, #$100	WC		' carry from bit 7 to bit 8?
		muxc	F, #CF
		and	alu, #$ff	WZ, WC		' zero and parity of result?
		muxz	F, #ZF
		muxnc	F, #PF
		test	alu, #$80	WC		' sign of result?
		muxc	F, #SF
szhpc_flags_ret
		ret

'***************************************************************************************************************
'
' Compute sign, zero and parity flags after logic/shift/rotate instruction
'
szp_flags
		muxz	F, #ZF				' if zero set Z flag
		muxnc	F, #PF				' if even parity set the P flag
		test	alu, #$80	WC		' get sign in prop's carry flag
		muxc	F, #SF				' set S flag
szp_flags_ret
		ret


'***************************************************************************************************************
'
' Add tmp to accumulator
'
alu_add
		test	F, #$100	WC		' clear carry flag
		jmp	#alu_adc_nc

'***************************************************************************************************************
'
' Add tmp with carry to accumulator
'
alu_adc
		test	F, #CF		WC		' get the C flag into carry
alu_adc_nc
		mov	tmp, alu			' 2nd operand to tmp
		mov	F, #0				' clear all flags
		mov	alu, A
		mov	aux, A
		addx	alu, tmp			' add the tmp value with carry
		movs	szhpc_flags_ret, #alu_to_a
		jmp	#szhpc_flags

'***************************************************************************************************************
'
' Subtract tmp from accumulator
'
alu_sub
		test	F, #$100	WC		' clear carry flag
		jmp	#alu_sbb_nc

'***************************************************************************************************************
'
' Subtract tmp with carry from accumulator
'
alu_sbb
		test	F, #CF		WC		' get the C flag into carry
alu_sbb_nc
		mov	tmp, alu			' 2nd operand to tmp
		mov	F, #NF				' clear all flags but the N flag
		mov	alu, A
		mov	aux, A
		subx	alu, tmp			' subtract the tmp value with carry
		movs	szhpc_flags_ret, #alu_to_a
		jmp	#szhpc_flags

'***************************************************************************************************************
'
' Logical and accumulator with tmp
'
alu_ana
		mov	tmp, alu			' 2nd operand to tmp
		mov	F, #HF				' clear all flags but H flag
		mov	alu, A
		and	alu, tmp	WZ, WC		' and with the tmp value
		movs	szp_flags_ret, #alu_to_a
		jmp	#szp_flags

'***************************************************************************************************************
'
' Logical or accumulator with tmp
'
' Note: removing the bits in alu from A will turn the following XOR into an OR
alu_ora
		andn	A, alu

'***************************************************************************************************************
'
' Logical exclusive or accumulator with tmp
'
alu_xra
		mov	tmp, alu			' 2nd operand to tmp
		mov	F, #0				' clear all flags
		mov	alu, A
		xor	alu, tmp	WZ, WC		' xor with the tmp value
		movs	szp_flags_ret, #alu_to_a
		jmp	#szp_flags

'***************************************************************************************************************
'
' Compare accumulator with tmp
'
alu_cmp
		mov	tmp, alu			' 2nd operand to tmp
		mov	F, #NF				' clear all flags but the N flag
		mov	alu, A
		mov	aux, A
		sub	alu, tmp			' subtract the tmp value
		movs	szhpc_flags_ret, #fetch
		jmp	#szhpc_flags

'***************************************************************************************************************
'
' Double add (16 bits) to HL register pair the value in alu
'
alu_dad
		mov	tmp, alu
		and	tmp, #$ff
		shr	alu, #8
		andn	F, #NF
		add	L, tmp
		test	L, #$100	WC
		and	L, #$ff
		mov	aux, H
		addx	H, alu
		test	H, #$100	WC
		muxc	F, #CF
		and	H, #$ff
		xor	aux, H
		xor	aux, alu
		test	aux, #HF	WC
		muxc	F, #HF
		jmp	#fetch

'******************************************************************************
'
' Rotate left circular
'
i_rlc		' 07
		shl	A, #1
		test	A, #$100	WC		' check carry
		and	A, #$ff				' keep bits 7..0
		muxc	A, #$01				' set low bit from carry
		and	F, #(SF | ZF | PF)		' keep S, Z and P flags
		muxc	F, #CF				' set C flag from carry
		jmp	#fetch

'******************************************************************************
'
' Rotate right circular
'
i_rrc		' 0f
		shr	A, #1		WC		' rotate right and check carry
		muxc	A, #$80				' set bit #7 from carry flag
		and	F, #(SF | ZF | PF)		' keep S, Z and P flags
		muxc	F, #CF				' set C from carry flag
		jmp	#fetch

'******************************************************************************
'
' Rotate left through carry
'
i_rl		' 17
		shl	A, #1
		test	A, #$100	WZ		' test bit #7
		and	A, #$ff				' keep bits 7..0
		test	F, #CF		WC		' get C flag to carry
		muxnz	F, #CF				' set C from zero flag
		muxc	A, #$01				' set bit #0 from previous C flag
		and	F, #(SF | ZF | PF | CF)		' keep S, Z, P and C flags
		jmp	#fetch

'******************************************************************************
'
' Rotate right through carry
'
i_rr		' 1f
		shr	A, #1		WC		' rotate right and check carry
		test	F, #CF		WZ		' get C flag
		muxc	F, #CF				' set C flag from carry flag
		muxnz	A, #$80				' set bit #7 from zero flag
		and	F, #(SF | ZF | PF | CF)		' keep S, Z, P and C flags
		jmp	#fetch

'******************************************************************************
'
' Complement accumulator
'
i_cma		' 2f
		xor	A, #$ff
		or	F, #(HF | NF)			' set the H and N flags
		jmp	#fetch

'******************************************************************************
'
' Set carry flag
'
i_stc		' 37
		or	F, #CF				' set the C flag
		andn	F, #(HF | NF)			' clear the H and N flags
		jmp	#fetch

'******************************************************************************
'
' Complement carry flag
'
i_cmc		' 3f
		test	F, #CF		WC		' get the C flag into the carry
		muxc	F, #HF				' set the H flag if C was set
		xor	F, #CF				' toggle the C flag
		andn	F, #NF				' clear N flag
		jmp	#fetch

'******************************************************************************
'
' Halt CPU
'
i_halt
#ifdef INTERRUPTS
		mov	HALT, #1
		sub	PC, #1
		jmp	#fetch
#else
		call	#do_break
		jmp	#fetch
#endif

'******************************************************************************
'
' Disable interrupts
'
i_di		' f3
		andn	IM, #IM_IE
		jmp	#fetch

'******************************************************************************
'
' Enable interrupts
'
i_ei		' fb
		or	IM, #IM_IE
#ifdef INTERRUPTS
		jmp	#fetch_0
#else
		jmp	#fetch
#endif

'******************************************************************************
'
' Get the status of the CPU's serial input device (SID) pin
'
get_sid
		mov	t1, #io_cmd_get_sid
		shl	t1, #24
		wrlong	t1, io_command
:wait		rdlong	t1, io_command
		shr	t1, #24			WZ, NR
	if_nz	jmp	#:wait
		mov	alu, t1
		and	alu, #$ff
get_sid_ret
		ret

'******************************************************************************
'
' Set the status of the CPU's serial output device (SOD) pin
'
set_sod
		mov	t1, #io_cmd_set_sod
		shl	t1, #24
		or	t1, alu
		wrlong	t1, io_command
:wait		rdlong	t1, io_command
		shr	t1, #24			WZ, NR
	if_nz	jmp	#:wait
set_sod_ret
		ret

' ----------------------------------------------------------------------------------------------
' Store all registers in their hub RAM locations
' and send a break command to the io handler
'
do_break
		mov	ea, context_ptr			' destination in hub RAM
		movd	:loop, #C
		mov	t1, #10				' 10 registers: C, B, E, D, L, H, F, A, IM, IM_TRAP
:loop		wrbyte	0-0, ea
		add	:loop, d1
		add	ea, #1
		djnz	t1, #:loop
		wrword	SP, ea
		add	ea, #2
		wrword	PC, ea
		wrlong	io_break, io_command
:wait		rdlong	t1, io_command
		shr	t1, #24		WZ, NR
	if_nz	jmp	#:wait
do_break_ret
		ret

'***************************************************************************************************************
'
' Execute code in LMM mode
'
lmm
		shr	vector, #9
		mov	lmm_pc, vector
lmm_next
		rdlong	:op1, lmm_pc
		add	lmm_pc, #4
:op1		nop
		rdlong	:op2, lmm_pc
		add	lmm_pc, #4
:op2		nop
		rdlong	:op3, lmm_pc
		add	lmm_pc, #4
:op3		nop
		rdlong	:op4, lmm_pc
		add	lmm_pc, #4
:op4		nop
		jmp	#lmm_next
lmm_next_ret
lmm_ret
		ret

'***************************************************************************************************************
'***************************************************************************************************************

		fit	$1f0

		org	$1f0
__par		res	1				' read only
__cnt		res	1				' read only
__ina		res	1				' read only
__inb		res	1				' read only
opcode		long					' most recent opcode
__outa		res	1				' read/write
alu		long					' arithmetic logic unit (sort of :)
__outb		res	1				' read/write
__dira		res	1				' read/write - leave this alone
__dirb		res	1				' read/write - leave this alone
__ctra		res	1				' read/write
__ctrb		res	1				' read/write
ea		long					' effective address
daa_adj		long					' adjustment value for DAA
__frqa		res	1				' read/write
tmp		long					' second operand temp
__frqb		res	1				' read/write
t1		long					' temporary 1
aux		long					' auxiliary flag temp
__phsa		res	1				' read/write
vector		long					' opcode vector
__phsb		res	1				' read/write
__vcfg		res	1				' read/write - leave this alone
lmm_pc		long					' LMM program counter
__vscl		res	1				' read/write

'***************************************************************************************************************
'***************************************************************************************************************
' Don't change the order of the registers unless you also
' change the order of their cog RAM counterparts and the
' break function
C_reg		byte	0
B_reg		byte	0
E_reg		byte	0
D_reg		byte	0
L_reg		byte	0
H_reg		byte	0
F_reg		byte	0
A_reg		byte	0
IM_reg		byte	0
IM_TRAP_reg	byte	0
SP_reg		word	0
PC_reg		word	0
		byte	0,0,0,0,0,0,0,0

'***************************************************************************************************************
'
'	LMM code fragments
'
'***************************************************************************************************************

#ifdef INTERRUPTS
' ----------------------------------------------------------------------------------------------
' Interrupt handler
'
do_interrupt
		shl	tmp, #1			WC, NR
	if_c	add	lmm_pc, #4*(:trap - $ - 1)	' bit #31 is set?
		test	IM, #IM_IE		WZ	' other interrupts enabled?
	if_z	jmp	lmm_ret				' no, continue with next opcode

		testn	tmp, #$78		WZ	' is it not just a vector request?
	if_nz	add	lmm_pc, #4*(:irq_call - $ - 1)	' check for CALL or RST opcode
		test	IM, #IM_I75		WZ	' pending interrupt RST 7.5 request?
	if_z	test	tmp, #IM_I75		WZ	' new interrupt RST 7.5 request?
	if_z	add	lmm_pc, #4*(:not_RST75 - $ - 1)

		or	IM, #IM_I75			' RST 7.5 is positive edge triggered
		test	IM, #IM_M75		WZ	' RST 7.5 masked?
	if_nz	add	lmm_pc, #4*(:not_RST75 - $ - 1)
		andn	tmp, #IM_I75			' clear IRQ for RST 7.5
		wrlong	tmp, irq_ptr			' (should be one in an interrupt acknowledge)
		xor	IM, #(IM_I75 | IM_M75)		' toggle IM_I75 off and IM_M75 on
		mov	alu, #ADDR_RST75
		add	lmm_pc, #4*(:take_int - $ - 1)

:not_RST75	test	tmp, #IM_I65		WZ	' interrupt RST 6.5 request?
		muxnz	IM, #IM_I65			' IM_I65 is level triggered
	if_z	add	lmm_pc, #4*(:not_RST65 - $ - 1)
		test	IM, #IM_M65		WZ	' RST 6.5 masked?
	if_nz	add	lmm_pc, #4*(:not_RST65 - $ - 1)
		andn	tmp, #IM_I65			' clear IRQ for RST 6.5
		wrlong	tmp, irq_ptr			' (should be one in an interrupt acknowledge)
		xor	IM, #(IM_I65 | IM_M65)		' toggle IM_I65 off and IM_M65 on
		mov	alu, #ADDR_RST65
		add	lmm_pc, #4*(:take_int - $ - 1)

:not_RST65	test	tmp, #IM_I55		WZ	' interrupt RST 5.5 request?
		muxnz	IM, #IM_I55
	if_z	add	lmm_pc, #4*(:not_RST55 - $ - 1)
		test	IM, #IM_M55		WZ	' RST 5.5 masked?
	if_nz	add	lmm_pc, #4*(:not_RST55 - $ - 1)
		andn	tmp, #IM_I55			' clear IRQ for RST 5.5
		wrlong	tmp, irq_ptr			' (should be one in an interrupt acknowledge)
		xor	IM, #(IM_I55 | IM_M55)		' toggle IM_I55 off and IM_M55 on
		mov	alu, #ADDR_RST55
		add	lmm_pc, #4*(:take_int - $ - 1)

:not_RST55	test	tmp, #IM_IE		WZ	' interrupt request?
	if_z	jmp	lmm_ret				' no, continue with next opcode
		' vectored interrupt request
		' This has to be changed if other vectors are required
		mov	alu, ADDR_INTR
		add	lmm_pc, #4*(:take_int - $ - 1)

:irq_call	mov	alu, tmp
		and	alu, #$ff			' mask LSB
		cmp	alu, #$cd		WZ	' is it a call opcode?
	if_z	shr	tmp, #8				' yes, call address in tmp
	if_nz	and	tmp, #$38			' no, mask RST vector in tmp
		mov	alu, tmp			' set as interrupt address
		mov	tmp, #0
		wrlong	tmp, irq_ptr			' (should be one in an interrupt acknowledge)
		add	lmm_pc, #4*(:take_int - $ - 1)

:trap		mov	IM_TRAP, IM			' remember IEN state in TRAP
		or	IM_TRAP, #$80
		mov	alu, #ADDR_TRAP			' trap vector
:take_int	andn	IM, #IM_IE			' disable further interrupts
		cmp	HALT, #0		WZ	' is the CPU halted?
		mov	HALT, #0			' clear the HALT flag
	if_nz	add	PC, #1				' exit from the HALT loop
		xor	alu, PC				' swap alu and PC
		xor	PC, alu
		xor	alu, PC
		call	#push_alu
		jmp	lmm_ret				' continue after the call to do_interrupt
#endif INTERRUPTS

' ----------------------------------------------------------------------------------------------
i_dsub	' 08	DSUB	B
		mov	t1, C
		mov	tmp, B
		sub	L, t1				' add/subtract the LSB tmp value with carry
		test	L, #$100	WC		' get LSB carry
		and	L, #$ff
		mov	F, #NF				' clear all flags (but N flag for SBC)
		mov	alu, H
		mov	aux, H
		subx	alu, tmp			' add/subtract the MSB tmp value with carry
		call	#szhpc_flags
	if_z	cmp	L, #0		WZ
	if_nz	andn	F, #ZF				' clear Z flag if L is non zero
		mov	H, alu
		jmp	#fetch

' ----------------------------------------------------------------------------------------------
i_asrh	' 10	ASRH
		'	HL := HL >> 1, CF = bit#0
		shr	H, #1			WC
		rcr	L, #1			WC
		muxc	F, #CF
		jmp	#fetch

' ----------------------------------------------------------------------------------------------
i_rlde		' 18	RLDE
		'	DE := DE rol 1, CF = bit#15, VF = bit#15 changes
		mov	aux, D
		shl	E, #1
		test	E, #$100		WC
		rcl	D, #1
		test	D, #$100		WC
		muxc	E, #1
		muxc	F, #CF
		xor	aux, D
		test	aux, #$80		WC
		muxc	F, #VF
		and	E, #$ff
		and	D, #$ff
		jmp	#fetch

' ----------------------------------------------------------------------------------------------
i_rim	' 20	RIM
		call	#get_sid			' get the SID status
		and	alu, alu		WZ
		muxnz	IM, #IM_SID			' set SID bit
		mov	A, IM
		andn	A, #(IM_I65 | IM_I55)		' copy live RST 6.5 and RST 5.5 states
		and	tmp, #(IM_I65 | IM_I55)		' mask the current requests for RST 6.5 and RST 5.5
		or	A, tmp
		mov	tmp, IM_TRAP			' get most recent TRAP interrupt mode
		test	tmp, #$80		WZ	' do we have a remembered IEN state?
	if_nz	andn	A, #IM_IE			' yes, clear current IEN
	if_nz	and	tmp, #IM_IE			' mask TRAP time IEN
	if_nz	or	A, tmp				' set IEN state at TRAP time
		mov	IM_TRAP, #0			' clear remembered IM state
		jmp	#fetch

' ----------------------------------------------------------------------------------------------
i_ldeh_imm8	' 28	LDEH	nn
		'	DE := HL + nn
		call	#rd_opcode
		mov	E, alu
		mov	D, H
		add	E, L
		test	E, #$100		WC
		and	E, #$ff
		addx	D, #0
		and	D, #$ff
		jmp	#fetch

' ----------------------------------------------------------------------------------------------
i_sim	' 30	SIM
		' set interrupt masks
		' bit 3 = 1: bits 2..0 define IM5.5, IM6.5 and IM7.5
		'	if IM5.5 == 0 then if irq[5.5] then IM |= I5.5
		'	if IM6.5 == 0 then if irq[6.5] then IM |= I6.5
		' bit 4 = 1: IM &= ~I7.5
		' bit 6 = 1: bit 7 is new SOD state
		test	A, #%00001000		WZ
	if_z	add	lmm_pc, #4*(:no_change - $ - 1)
		andn	IM, #(IM_M55 | IM_M65 | IM_M75 | IM_I55 | IM_I65)
		mov	alu, A
		and	alu, #(IM_M55 | IM_M65 | IM_M75)
		or	IM, alu
		rdlong	tmp, irq_ptr			' fetch pending IRQs
		and	tmp, #(IM_I55 | IM_I65)		' mask RST 6.5 and RST 5.5 requests
		or	IM, tmp
:no_change	test	A, #%00010000		WZ	' if bit 4 is set, IM_I75 is cleared
	if_nz	andn	IM, #IM_I75
		test	A, #%01000000		WZ	' if bit 6 is set, bit 7 is new SOD state
	if_nz	mov	alu, A
	if_nz	shr	alu, #7
	if_nz	call	#set_sod
		jmp	#fetch

' ----------------------------------------------------------------------------------------------
i_ldes_imm8	' 38	LDES	nn
		'	DE := SP + nn
		call	#rd_opcode
		mov	E, SP
		mov	D, SP
		shr	D, #8
		and	E, #$ff
		add	E, alu
		test	E, #$100		WC
		and	E, #$ff
		addx	D, #0
		and	D, #$ff
		jmp	#fetch

' ----------------------------------------------------------------------------------------------
i_daa		' 27
		mov	alu, A
		mov	tmp, alu
		mov	aux, alu			' save for auxiliary flag test
		and	tmp, #15			' isolate low nibble of accumulator
		mov	daa_adj, #$06			' +$06
		test	F, #NF		WZ		' test negative flag
		negnz	daa_adj, daa_adj		' -$06
		test	F, #HF		WZ
		cmp	tmp, #$09+1	WC		' aux isn't set; is nibble > $9 ?
  if_nc_or_nz	add	A, daa_adj			' aux was set or nibble is > $9 => add $06 or -$06
		shl	daa_adj, #4			' +$60 or -$60
		test	F, #CF		WZ
		cmp	alu, #$99+1	WC		' carry isn't set; is data > $99 ?
  if_nc_or_nz	add	A, daa_adj			' carry was set or data is > $99 => add $60 or -$60
		and	F, #(NF | CF)			' preserve old CF and NF
		cmp	alu, #$99+1	WC
	if_nc	or	F, #CF				' set carry if result > $99
		xor	aux, A
		and	aux, #HF	WZ		' if bit #4 of result changed
		or	F, aux				' set the auxiliary carry flag
		and	A, #$ff		WZ, WC		' get zero and parity flags
		muxz	F, #ZF				' set Z on zero
		muxnc	F, #PF				' clear P on parity
		test	A, #$80		WZ		' get sign bit #7
		muxnz	F, #SF				' set S on sign
		jmp	#fetch

' ----------------------------------------------------------------------------------------------
i_shlx_d	' d9	SHLX
		'	(DE) := HL
		mov	ea, D
		shl	ea, #8
		or	ea, E
		mov	alu, L
		call	#wr_byte
		add	ea, #1
		mov	alu, H
		call	#wr_byte
		jmp	#fetch

' ----------------------------------------------------------------------------------------------
i_rst_v		' cb	RST	V
		'	RST 40h if VF is set
		test	F, #VF
	if_nz	mov	alu, #ADDR_RSTV
	if_nz	jmp	#do_call
		jmp	#fetch

' ----------------------------------------------------------------------------------------------
i_xthl	' e3
		mov	tmp, H
		shl	tmp, #8
		or	tmp, L
		mov	ea, SP
		call	#rd_word
		xor	alu, tmp
		xor	tmp, alu
		xor	alu, tmp
		mov	ea, SP
		call	#wr_word
		mov	L, tmp
		and	L, #$ff
		mov	H, tmp
		shr	H, #8
		jmp	#fetch

' ----------------------------------------------------------------------------------------------
i_xchg		' eb
		xor	L, E
		xor	E, L
		xor	L, E
		xor	H, D
		xor	D, H
		xor	H, D
		jmp	#fetch

' ----------------------------------------------------------------------------------------------
i_lhlx_d	' ed	LHLX
		'	HL := (DE)
		mov	ea, D
		shl	ea, #8
		or	ea, E
		call	#rd_byte
		mov	L, alu
		add	ea, #1
		call	#rd_byte
		mov	H, alu
		jmp	#fetch
' ----------------------------------------------------------------------------------------------


opcodes_00
	{00}	long	fetch								' NOP
	{01}	long	imm16_to_alu	| alu_to_bc        << 9				' LXI  B,nnnn
	{02}	long	bc_to_ea	| a_to_alu         << 9	| wr_byte_ea      << 18	' STAX B
	{03}	long	bc_to_alu	| inc16	           << 9	| alu_to_bc       << 18	' INX  B 
	{04}	long	b_to_alu	| alu_inc          << 9	| alu_to_b        << 18	' INR  B
	{05}	long	b_to_alu	| alu_dec          << 9	| alu_to_b        << 18	' DCR  B
	{06}	long	imm8_to_alu	| alu_to_b         << 9				' MVI  B,nn
	{07}	long	i_rlc								' RLC

	{08}	long	lmm		| @@@i_dsub        << 9				' DSUB B    (8085)
	{09}	long	bc_to_alu	| alu_dad          << 9				' DAD  B
	{0a}	long	bc_to_ea	| rd_byte_ea       << 9	| alu_to_a        << 18	' LDAX B
	{0b}	long	bc_to_alu	| dec16	           << 9	| alu_to_bc       << 18	' DCX  B
	{0c}	long	c_to_alu	| alu_inc          << 9	| alu_to_c        << 18	' INR  C
	{0d}	long	c_to_alu	| alu_dec          << 9	| alu_to_c        << 18	' DCR  C
	{0e}	long	imm8_to_alu	| alu_to_c         << 9				' MVI  C,nn
	{0f}	long	i_rrc								' RRC

	{10}	long	lmm		| @@@i_asrh					' ASRH      (8085)
	{11}	long	imm16_to_alu	| alu_to_de        << 9				' LXI  D,nnnn
	{12}	long	de_to_ea	| a_to_alu         << 9	| wr_byte_ea      << 18	' STAX D
	{13}	long	de_to_alu	| inc16	           << 9	| alu_to_de       << 18	' INX  D 
	{14}	long	d_to_alu	| alu_inc          << 9	| alu_to_d        << 18	' INR  D
	{15}	long	d_to_alu	| alu_dec          << 9	| alu_to_d        << 18	' DCR  D
	{16}	long	imm8_to_alu	| alu_to_d         << 9				' MVI  D,nn
	{17}	long	i_rl								' RL

	{18}	long	lmm		| @@@i_rlde        << 9				' RLDE      (8085)
	{19}	long	de_to_alu	| alu_dad          << 9				' DAD  D
	{1a}	long	de_to_ea	| rd_byte_ea       << 9	| alu_to_a        << 18	' LDAX D
	{1b}	long	de_to_alu	| dec16	           << 9	| alu_to_de       << 18	' DCX  D
	{1c}	long	e_to_alu	| alu_inc          << 9	| alu_to_e        << 18	' INR  E
	{1d}	long	e_to_alu	| alu_dec          << 9	| alu_to_e        << 18	' DCR  E
	{1e}	long	imm8_to_alu	| alu_to_e         << 9				' MVI  E,nn
	{1f}	long	i_rr								' RR

	{20}	long	lmm		| @@@i_rim         << 9				' RIM       (8085)
	{21}	long	imm16_to_alu	| alu_to_hl        << 9				' LXI  H,nnnn
	{22}	long	imm16_to_ea	| hl_to_alu        << 9	| wr_word_ea      << 18	' SHLX nnnn
	{23}	long	hl_to_alu	| inc16	           << 9	| alu_to_hl       << 18	' INX  H 
	{24}	long	h_to_alu	| alu_inc          << 9	| alu_to_h        << 18	' INR  H
	{25}	long	h_to_alu	| alu_dec          << 9	| alu_to_h        << 18	' DCR  H
	{26}	long	imm8_to_alu	| alu_to_h         << 9				' MVI  H,nn
	{27}	long	lmm		| @@@i_daa         << 9				' DAA

	{28}	long	lmm		| @@@i_ldeh_imm8   << 9				' LDEH nn   (8085)
	{29}	long	hl_to_alu	| alu_dad          << 9				' DAD  H
	{2a}	long	imm16_to_ea	| rd_word_ea       << 9	| alu_to_hl       << 18	' LHLX nnnn
	{2b}	long	hl_to_alu	| dec16	           << 9	| alu_to_hl       << 18	' DCX  H
	{2c}	long	l_to_alu	| alu_inc          << 9	| alu_to_l        << 18	' INR  L
	{2d}	long	l_to_alu	| alu_dec          << 9	| alu_to_l        << 18	' DCR  L
	{2e}	long	imm8_to_alu	| alu_to_l         << 9				' MVI  L,nn
	{2f}	long	i_cma								' CMA

	{30}	long	lmm		| @@@i_sim         << 9				' SIM       (8085)
	{31}	long	imm16_to_alu	| alu_to_sp        << 9				' LXI  SP,nnnn
	{32}	long	imm16_to_ea	| a_to_alu         << 9	| wr_byte_ea      << 18	' STAX nnnn
	{33}	long	sp_to_alu	| inc16	           << 9	| alu_to_sp       << 18	' INX  SP
	{34}	long	m_to_alu	| alu_inc          << 9	| alu_to_m        << 18	' INR  M
	{35}	long	m_to_alu	| alu_dec          << 9	| alu_to_m        << 18	' DCR  M
	{36}	long	imm8_to_alu	| hl_to_ea         << 9 | alu_to_m        << 18	' MVI  M,nn
	{37}	long	i_stc								' STC

	{38}	long	lmm		| @@@i_ldes_imm8   << 9				' LDES nn   (8085)
	{39}	long	sp_to_alu	| alu_dad          << 9				' DAD  SP
	{3a}	long	imm16_to_ea	| rd_byte_ea       << 9	| alu_to_a        << 18	' LDAX nnnn
	{3b}	long	sp_to_alu	| dec16	           << 9	| alu_to_sp       << 18	' DCX  SP
	{3c}	long	a_to_alu	| alu_inc          << 9	| alu_to_a        << 18	' INR  A
	{3d}	long	a_to_alu	| alu_dec          << 9	| alu_to_a        << 18	' DCR  A
	{3e}	long	imm8_to_alu	| alu_to_a         << 9				' MVI  A,nn
	{3f}	long	i_cmc								' CMC

	{40}	long	fetch								' MOV  B,B
	{41}	long	c_to_alu	| alu_to_b         << 9				' MOV  B,C
	{42}	long	d_to_alu	| alu_to_b         << 9				' MOV  B,D
	{43}	long	e_to_alu	| alu_to_b         << 9				' MOV  B,E
	{44}	long	h_to_alu	| alu_to_b         << 9				' MOV  B,H
	{45}	long	l_to_alu	| alu_to_b         << 9				' MOV  B,L
	{46}	long	m_to_alu	| alu_to_b         << 9				' MOV  B,M
	{47}	long	a_to_alu	| alu_to_b         << 9				' MOV  B,A

	{48}	long	b_to_alu	| alu_to_c         << 9				' MOV  C,B
	{49}	long	fetch								' MOV  C,C
	{4a}	long	d_to_alu	| alu_to_c         << 9				' MOV  C,D
	{4b}	long	e_to_alu	| alu_to_c         << 9				' MOV  C,E
	{4c}	long	h_to_alu	| alu_to_c         << 9				' MOV  C,H
	{4d}	long	l_to_alu	| alu_to_c         << 9				' MOV  C,L
	{4e}	long	m_to_alu	| alu_to_c         << 9				' MOV  C,M
	{4f}	long	a_to_alu	| alu_to_c         << 9				' MOV  C,A

	{50}	long	b_to_alu	| alu_to_d         << 9				' MOV  D,B
	{51}	long	c_to_alu	| alu_to_d         << 9				' MOV  D,C
	{52}	long	fetch								' MOV  D,D
	{53}	long	e_to_alu	| alu_to_d         << 9				' MOV  D,E
	{54}	long	h_to_alu	| alu_to_d         << 9				' MOV  D,H
	{55}	long	l_to_alu	| alu_to_d         << 9				' MOV  D,L
	{56}	long	m_to_alu	| alu_to_d         << 9				' MOV  D,M
	{57}	long	a_to_alu	| alu_to_d         << 9				' MOV  D,A

	{58}	long	b_to_alu	| alu_to_e         << 9				' MOV  E,B
	{59}	long	c_to_alu	| alu_to_e         << 9				' MOV  E,C
	{5a}	long	d_to_alu	| alu_to_e         << 9				' MOV  E,D
	{5b}	long	fetch								' MOV  E,E
	{5c}	long	h_to_alu	| alu_to_e         << 9				' MOV  E,H
	{5d}	long	l_to_alu	| alu_to_e         << 9				' MOV  E,L
	{5e}	long	m_to_alu	| alu_to_e         << 9				' MOV  E,M
	{5f}	long	a_to_alu	| alu_to_e         << 9				' MOV  E,A

	{60}	long	b_to_alu	| alu_to_h         << 9				' MOV  H,B
	{61}	long	c_to_alu	| alu_to_h         << 9				' MOV  H,C
	{62}	long	d_to_alu	| alu_to_h         << 9				' MOV  H,D
	{63}	long	e_to_alu	| alu_to_h         << 9				' MOV  H,E
	{64}	long	fetch								' MOV  H,H
	{65}	long	l_to_alu	| alu_to_h         << 9				' MOV  H,L
	{66}	long	m_to_alu	| alu_to_h         << 9				' MOV  H,M
	{67}	long	a_to_alu	| alu_to_h         << 9				' MOV  H,A

	{68}	long	b_to_alu	| alu_to_l         << 9				' MOV  L,B
	{69}	long	c_to_alu	| alu_to_l         << 9				' MOV  L,C
	{6a}	long	d_to_alu	| alu_to_l         << 9				' MOV  L,D
	{6b}	long	e_to_alu	| alu_to_l         << 9				' MOV  L,E
	{6c}	long	h_to_alu	| alu_to_l         << 9				' MOV  L,H
	{6d}	long	fetch								' MOV  L,L
	{6e}	long	m_to_alu	| alu_to_l         << 9				' MOV  L,M
	{6f}	long	a_to_alu	| alu_to_l         << 9				' MOV  L,A

	{70}	long	hl_to_ea	| b_to_alu         << 9	| alu_to_m        << 18	' MOV  M,B
	{70}	long	hl_to_ea	| c_to_alu         << 9	| alu_to_m        << 18	' MOV  M,C
	{70}	long	hl_to_ea	| d_to_alu         << 9	| alu_to_m        << 18	' MOV  M,D
	{70}	long	hl_to_ea	| e_to_alu         << 9	| alu_to_m        << 18	' MOV  M,E
	{70}	long	hl_to_ea	| h_to_alu         << 9	| alu_to_m        << 18	' MOV  M,H
	{70}	long	hl_to_ea	| l_to_alu         << 9	| alu_to_m        << 18	' MOV  M,L
	{76}	long	i_halt								' HALT
	{70}	long	hl_to_ea	| a_to_alu         << 9	| alu_to_m        << 18	' MOV  M,A

	{78}	long	b_to_alu	| alu_to_a         << 9				' MOV  A,B
	{79}	long	c_to_alu	| alu_to_a         << 9				' MOV  A,C
	{7a}	long	d_to_alu	| alu_to_a         << 9				' MOV  A,D
	{7b}	long	e_to_alu	| alu_to_a         << 9				' MOV  A,E
	{7c}	long	h_to_alu	| alu_to_a         << 9				' MOV  A,H
	{7d}	long	l_to_alu	| alu_to_a         << 9				' MOV  A,L
	{7e}	long	m_to_alu	| alu_to_a         << 9				' MOV  A,M
	{7f}	long	fetch								' MOV  A,A

	{80}	long	b_to_alu	| alu_add          << 9				' ADD  B
	{81}	long	c_to_alu	| alu_add          << 9				' ADD  C
	{82}	long	d_to_alu	| alu_add          << 9				' ADD  D
	{83}	long	e_to_alu	| alu_add          << 9				' ADD  E
	{84}	long	h_to_alu	| alu_add          << 9				' ADD  H
	{85}	long	l_to_alu	| alu_add          << 9				' ADD  L
	{86}	long	m_to_alu	| alu_add          << 9				' ADD  M
	{87}	long	a_to_alu	| alu_add          << 9				' ADD  A

	{88}	long	b_to_alu	| alu_adc          << 9				' ADC  B
	{89}	long	c_to_alu	| alu_adc          << 9				' ADC  C
	{8a}	long	d_to_alu	| alu_adc          << 9				' ADC  D
	{8b}	long	e_to_alu	| alu_adc          << 9				' ADC  E
	{8c}	long	h_to_alu	| alu_adc          << 9				' ADC  H
	{8d}	long	l_to_alu	| alu_adc          << 9				' ADC  L
	{8e}	long	m_to_alu	| alu_adc          << 9				' ADC  M
	{8f}	long	a_to_alu	| alu_adc          << 9				' ADC  A

	{90}	long	b_to_alu	| alu_sub          << 9				' SUB  B
	{91}	long	c_to_alu	| alu_sub          << 9				' SUB  C
	{92}	long	d_to_alu	| alu_sub          << 9				' SUB  D
	{93}	long	e_to_alu	| alu_sub          << 9				' SUB  E
	{94}	long	h_to_alu	| alu_sub          << 9				' SUB  H
	{95}	long	l_to_alu	| alu_sub          << 9				' SUB  L
	{96}	long	m_to_alu	| alu_sub          << 9				' SUB  M
	{97}	long	a_to_alu	| alu_sub          << 9				' SUB  A

	{98}	long	b_to_alu	| alu_sbb          << 9				' SBB  B
	{99}	long	c_to_alu	| alu_sbb          << 9				' SBB  C
	{9a}	long	d_to_alu	| alu_sbb          << 9				' SBB  D
	{9b}	long	e_to_alu	| alu_sbb          << 9				' SBB  E
	{9c}	long	h_to_alu	| alu_sbb          << 9				' SBB  H
	{9d}	long	l_to_alu	| alu_sbb          << 9				' SBB  L
	{9e}	long	m_to_alu	| alu_sbb          << 9				' SBB  M
	{9f}	long	a_to_alu	| alu_sbb          << 9				' SBB  A

	{a0}	long	b_to_alu	| alu_ana          << 9				' ANA  B
	{a1}	long	c_to_alu	| alu_ana          << 9				' ANA  C
	{a2}	long	d_to_alu	| alu_ana          << 9				' ANA  D
	{a3}	long	e_to_alu	| alu_ana          << 9				' ANA  E
	{a4}	long	h_to_alu	| alu_ana          << 9				' ANA  H
	{a5}	long	l_to_alu	| alu_ana          << 9				' ANA  L
	{a6}	long	m_to_alu	| alu_ana          << 9				' ANA  M
	{a7}	long	a_to_alu	| alu_ana          << 9				' ANA  A

	{a8}	long	b_to_alu	| alu_xra          << 9				' XRA  B
	{a9}	long	c_to_alu	| alu_xra          << 9				' XRA  C
	{aa}	long	d_to_alu	| alu_xra          << 9				' XRA  D
	{ab}	long	e_to_alu	| alu_xra          << 9				' XRA  E
	{ac}	long	h_to_alu	| alu_xra          << 9				' XRA  H
	{ad}	long	l_to_alu	| alu_xra          << 9				' XRA  L
	{ae}	long	m_to_alu	| alu_xra          << 9				' XRA  M
	{af}	long	a_to_alu	| alu_xra          << 9				' XRA  A

	{b0}	long	b_to_alu	| alu_ora          << 9				' ORA  B
	{b1}	long	c_to_alu	| alu_ora          << 9				' ORA  C
	{b2}	long	d_to_alu	| alu_ora          << 9				' ORA  D
	{b3}	long	e_to_alu	| alu_ora          << 9				' ORA  E
	{b4}	long	h_to_alu	| alu_ora          << 9				' ORA  H
	{b5}	long	l_to_alu	| alu_ora          << 9				' ORA  L
	{b6}	long	m_to_alu	| alu_ora          << 9				' ORA  M
	{b7}	long	a_to_alu	| alu_ora          << 9				' ORA  A
 
	{b8}	long	b_to_alu	| alu_cmp          << 9				' CMP  B
	{b9}	long	c_to_alu	| alu_cmp          << 9				' CMP  C
	{ba}	long	d_to_alu	| alu_cmp          << 9				' CMP  D
	{bb}	long	e_to_alu	| alu_cmp          << 9				' CMP  E
	{bc}	long	h_to_alu	| alu_cmp          << 9				' CMP  H
	{bd}	long	l_to_alu	| alu_cmp          << 9				' CMP  L
	{be}	long	m_to_alu	| alu_cmp          << 9				' CMP  M
	{bf}	long	a_to_alu	| alu_cmp          << 9				' CMP  A

	{c0}	long	test_flag_0	| ret_cond         << 9 | ZF              << 18	' RNZ
	{c1}	long	pop_bc								' POP  B
	{c2}	long	test_flag_0	| jmp_cond         << 9 | ZF              << 18	' JNZ  nnnn
	{c3}	long	imm16_to_alu	| alu_to_pc        << 9				' JMP  nnnn
	{c4}	long	test_flag_0	| call_cond        << 9 | ZF              << 18	' CNZ  nnnn
	{c5}	long	bc_to_alu	| push_reg         << 9				' PUSH B
	{c6}	long	imm8_to_alu	| alu_add          << 9				' ADI  imm8
	{c7}	long	rst								' RST  0

	{c8}	long	test_flag_1	| ret_cond         << 9 | ZF              << 18	' RZ
	{c9}	long	do_ret								' RET
	{ca}	long	test_flag_1	| jmp_cond         << 9 | ZF              << 18	' JZ   nnnn
	{cb}	long	lmm		| @@@i_rst_v       << 9				' RST  V      (8085)
	{cc}	long	test_flag_1	| call_cond        << 9 | ZF              << 18	' CZ   nnnn
	{cd}	long	imm16_to_alu	| do_call          << 9				' CALL nnnn
	{ce}	long	imm8_to_alu	| alu_adc          << 9				' ACI  imm8
	{cf}	long	rst								' RST  1

	{d0}	long	test_flag_0	| ret_cond         << 9 | CF              << 18	' RNC
	{c1}	long	pop_de								' POP  D
	{d2}	long	test_flag_0	| jmp_cond         << 9 | CF              << 18	' JNC  nnnn
	{d3}	long	a_to_port							' OUT  nn
	{d4}	long	test_flag_0	| call_cond        << 9 | CF              << 18	' CNC  nnnn
	{d5}	long	de_to_alu	| push_reg         << 9				' PUSH D
	{d6}	long	imm8_to_alu	| alu_sub          << 9				' SUI  imm8
	{d7}	long	rst								' RST  2

	{d8}	long	test_flag_1	| ret_cond         << 9 | CF              << 18	' RC
	{d9}	long	lmm		| @@@i_shlx_d      << 9				' SHLX D      (8085)
	{da}	long	test_flag_1	| jmp_cond         << 9 | CF              << 18	' JC   nnnn
	{db}	long	port_to_a							' IN   nn
	{dc}	long	test_flag_1	| call_cond        << 9 | CF              << 18	' CC   nnnn
	{dd}	long	test_flag_0	| jmp_cond         << 9	| X5              << 18	' JNX  nnnn   (8085)
	{de}	long	imm8_to_alu	| alu_sbb          << 9				' SBI  imm8
	{df}	long	rst								' RST  3

	{e0}	long	test_flag_0	| ret_cond         << 9 | PF              << 18	' RPE
	{e1}	long	pop_hl								' POP  H
	{e2}	long	test_flag_0	| jmp_cond         << 9 | PF              << 18	' JPE  nnnn
	{e3}	long	lmm		| @@@i_xthl        << 9				' XTHL
	{e4}	long	test_flag_0	| call_cond        << 9 | PF              << 18	' CPE  nnnn
	{e5}	long	hl_to_alu	| push_reg         << 9				' PUSH H
	{e6}	long	imm8_to_alu	| alu_ana          << 9				' ANI  imm8
	{e7}	long	rst								' RST  4

	{e8}	long	test_flag_1	| ret_cond         << 9 | PF              << 18	' RPO
	{e9}	long	hl_to_alu	| alu_to_pc        << 9				' PCHL
	{ea}	long	test_flag_1	| jmp_cond         << 9 | PF              << 18	' JPO  nnnn
	{eb}	long	lmm		| @@@i_xchg        << 9				' XCHG
	{ec}	long	test_flag_1	| call_cond        << 9 | PF              << 18	' CPO  nnnn
	{ed}	long	lmm		| @@@i_lhlx_d      << 9				' LHLX D
	{ee}	long	imm8_to_alu	| alu_xra          << 9				' XRI  imm8
	{ef}	long	rst								' RST  5

	{f0}	long	test_flag_0	| ret_cond         << 9 | SF              << 18	' RP
	{f1}	long	pop_af								' POP  PSW
	{f2}	long	test_flag_0	| jmp_cond         << 9 | SF              << 18	' JP   nnnn
	{f3}	long	i_di								' DI
	{f4}	long	test_flag_0	| call_cond        << 9 | SF              << 18	' CP   nnnn
	{f5}	long	af_to_alu	| push_reg         << 9				' PUSH PSW
	{f6}	long	imm8_to_alu	| alu_ora          << 9				' ORI  imm8
	{f7}	long	rst								' RST  6

	{f8}	long	test_flag_1	| ret_cond         << 9 | SF              << 18	' RM
	{f9}	long	hl_to_alu	| alu_to_sp        << 9				' SPHL
	{fa}	long	test_flag_1	| jmp_cond         << 9 | SF              << 18	' JM   nnnn
	{fb}	long	i_ei								' EI
	{fc}	long	test_flag_1	| call_cond        << 9 | SF              << 18	' CM   nnnn
	{fd}	long	test_flag_1	| jmp_cond         << 9	| X5              << 18	' JX   nnnn
	{fe}	long	imm8_to_alu	| alu_cmp          << 9				' CMI  imm8
	{ff}	long	rst								' RST  7

#ifdef PAGED_MEMORY

#ifdef GAMEBOY
rd_memory
	'	x000-x3ff x400-x7ff x800-xbff xc00-xfff
	byte	rd_rom, rd_rom, rd_rom, rd_rom	' 0000-0fff
	byte	rd_rom, rd_rom, rd_rom, rd_rom	' 1000-1fff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' 2000-2fff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' 3000-3fff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' 4000-4fff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' 5000-5fff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' 6000-6fff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' 7000-7fff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' 8000-8fff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' 9000-9fff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' a000-afff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' b000-bfff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' c000-cfff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' d000-dfff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' e000-efff
	byte	rd_nop, rd_nop, rd_nop, rd_nop	' f000-ffff

wr_memory
	'	x000-x3ff x400-x7ff x800-xbff xc00-xfff
	byte	wr_rom, wr_rom, wr_rom, wr_rom	' 0000-0fff
	byte	wr_rom, wr_rom, wr_rom, wr_rom	' 1000-1fff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' 2000-2fff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' 3000-3fff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' 4000-4fff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' 5000-5fff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' 6000-6fff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' 7000-7fff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' 8000-8fff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' 9000-9fff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' a000-afff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' b000-bfff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' c000-cfff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' d000-dfff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' e000-efff
	byte	wr_nop, wr_nop, wr_nop, wr_nop	' f000-ffff
#endif GAMEBOY

#endif PAGED_MEMORY
